from typing import List, Optional
from ddd_objects.domain.entity import Entity, ExpiredEntity
from .value_obj import (
    Port,
    Weight,
    InstanceType,
    Number,
    Status,
    Path,
    Password,
    GPUType,
    Name,
    Hostname,
    RecordID,
    InternetPayType,
    InstanceID,
    RegionID,
    Usage,
    BandWidth,
    Type,
    ID,
    IP,
    DNSLine,
    Value,
    SecurityGroupID,
    Output,
    DomainName,
    InstanceName,
    Price,
    ImageID,
    PayType,
    Size,
    Command,
    Subdomain,
    Endpoint,
    Data,
    Bool,
    Username,
    InstanceTypeStatusCategory,
    DateTime,
    DNSType,
    InstanceTypeStatus,
    ZoneID
)

class Condition(Entity):
    def __init__(
        self,
        min_cpu_num: Number,
        max_cpu_num: Number,
        min_memory_size: Size,
        max_memory_size: Size,
        min_gpu_num: Optional[Number] = None,
        max_gpu_num: Optional[Number] = None,
        min_gpu_memory_size: Optional[Size] = None,
        max_gpu_memory_size: Optional[Size] = None
    ):
        self.min_cpu_num=min_cpu_num
        self.max_cpu_num=max_cpu_num
        self.min_memory_size=min_memory_size
        self.max_memory_size=max_memory_size
        self.min_gpu_num=min_gpu_num
        self.max_gpu_num=max_gpu_num
        self.min_gpu_memory_size=min_gpu_memory_size
        self.max_gpu_memory_size=max_gpu_memory_size

class InstanceInfo(ExpiredEntity):
    def __init__(
        self,
        id: ID,
        status: Status,
        security_group_id: List[SecurityGroupID],
        instance_type: InstanceType,
        name: Name,
        hostname: Hostname,
        price: Price,
        image_id: ImageID,
        region_id: RegionID,
        zone_id: ZoneID,
        internet_pay_type: InternetPayType,
        pay_type: PayType,
        create_time: DateTime,
        os_name: Name,
        public_ip: List[IP],
        private_ip: IP,
        bandwidth_in: BandWidth,
        bandwidth_out: BandWidth,
        expired_time: DateTime,
        auto_release_time: DateTime,
        key_name: Name,
        _life_time: Number = Number(5)
    ):
        self.id=id
        self.status=status
        self.security_group_id=security_group_id
        self.instance_type=instance_type
        self.name=name
        self.hostname=hostname
        self.price=price
        self.image_id=image_id
        self.region_id=region_id
        self.zone_id=zone_id
        self.internet_pay_type=internet_pay_type
        self.pay_type=pay_type
        self.create_time=create_time
        self.os_name=os_name
        self.public_ip=public_ip
        self.private_ip=private_ip
        self.bandwidth_in=bandwidth_in
        self.bandwidth_out=bandwidth_out
        self.expired_time=expired_time
        self.auto_release_time=auto_release_time
        self.key_name=key_name
        self._life_time=_life_time
        super().__init__(_life_time)

class InstanceUserSetting(Entity):
    def __init__(
        self,
        name: Name,
        password: Password,
        image_id: ImageID,
        region_id: RegionID,
        exclude_instance_types: List[InstanceType],
        user_data: Optional[Data] = None,
        internet_pay_type: Type = Type('PayByTraffic'),
        amount: Number = Number(1),
        bandwidth_in: BandWidth = BandWidth(200),
        bandwidth_out: BandWidth = BandWidth(1),
        disk_size: Size = Size(20),
        key_name: Name = Name('ansible'),
        inner_connection: Bool = Bool(True)
    ):
        self.name=name
        self.password=password
        self.image_id=image_id
        self.region_id=region_id
        self.exclude_instance_types=exclude_instance_types
        self.user_data=user_data
        self.internet_pay_type=internet_pay_type
        self.amount=amount
        self.bandwidth_in=bandwidth_in
        self.bandwidth_out=bandwidth_out
        self.disk_size=disk_size
        self.key_name=key_name
        self.inner_connection=inner_connection

class CommandSetting(Entity):
    def __init__(
        self,
        command: Command = Command('echo 123'),
        forks: Number = Number(100),
        timeout: Number = Number(30),
        password: Password = None,
        username: Username = Username('root'),
        port: Port = Port(22),
        inner_connection: Bool = Bool(True)
    ):
        self.command=command
        self.forks=forks
        self.timeout=timeout
        self.password=password
        self.username=username
        self.port=port
        self.inner_connection=inner_connection

class CommandResult(Entity):
    def __init__(
        self,
        output: Output,
        instance_id: InstanceID,
        instance_name: InstanceName,
        ip: IP,
        succeed: Bool
    ):
        self.output=output
        self.instance_id=instance_id
        self.instance_name=instance_name
        self.ip=ip
        self.succeed=succeed

class OSSOperationInfo(Entity):
    def __init__(
        self,
        name: Name,
        endpoint: Endpoint,
        bucket_name: Name,
        local_path: Path,
        target_path: Path,
        with_tar: Bool = Bool(False)
    ):
        self.name=name
        self.endpoint=endpoint
        self.bucket_name=bucket_name
        self.local_path=local_path
        self.target_path=target_path
        self.with_tar=with_tar

class InstanceTypeWithStatus(ExpiredEntity):
    def __init__(
        self,
        region_id: RegionID,
        zone_id: ZoneID,
        instance_type_id: InstanceType,
        cpu_number: Number,
        memory_size: Size,
        gpu_type: GPUType,
        gpu_number: Number,
        status: InstanceTypeStatus,
        status_category: InstanceTypeStatusCategory,
        _life_time: Number = Number(5)
    ):
        self.region_id=region_id
        self.zone_id=zone_id
        self.instance_type_id=instance_type_id
        self.cpu_number=cpu_number
        self.memory_size=memory_size
        self.gpu_type=gpu_type
        self.gpu_number=gpu_number
        self.status=status
        self.status_category=status_category
        self._life_time=_life_time
        super().__init__(_life_time)

class InstanceTypeUserSetting(Entity):
    def __init__(
        self,
        region_id: RegionID,
        zone_id: ZoneID,
        instance_type_id: InstanceType
    ):
        self.region_id=region_id
        self.zone_id=zone_id
        self.instance_type_id=instance_type_id

class InstanceUsageInfo(ExpiredEntity):
    def __init__(
        self,
        instance_id: InstanceID,
        instance_name: InstanceName,
        cpu_number: Number,
        cpu_usage: Usage,
        memory_size: Size,
        memory_usage: Usage,
        flow_in: Size,
        flow_out: Size,
        disk_size: Size,
        disk_usage: Usage,
        io_in: Number,
        io_out: Number,
        _life_time: Number = Number(30)
    ):
        self.instance_id=instance_id
        self.instance_name=instance_name
        self.cpu_number=cpu_number
        self.cpu_usage=cpu_usage
        self.memory_size=memory_size
        self.memory_usage=memory_usage
        self.flow_in=flow_in
        self.flow_out=flow_out
        self.disk_size=disk_size
        self.disk_usage=disk_usage
        self.io_in=io_in
        self.io_out=io_out
        self._life_time=_life_time
        super().__init__(_life_time)

class DNSRecord(Entity):
    def __init__(
        self,
        domain_name: DomainName,
        subdomain: Subdomain,
        value: Value,
        id: Optional[RecordID] = None,
        weight: Optional[Weight] = None,
        dns_type: DNSType = DNSType('A'),
        ttl: Number = Number(600),
        priority: Optional[Number] = None,
        line: Optional[DNSLine] = None
    ):
        self.domain_name=domain_name
        self.subdomain=subdomain
        self.value=value
        self.id=id
        self.weight=weight
        self.dns_type=dns_type
        self.ttl=ttl
        self.priority=priority
        self.line=line

    def __eq__(self, __o: object) -> bool:
        if self.domain_name==__o.domain_name and self.subdomain==__o.subdomain \
            and self.value==__o.value  \
            and self.dns_type==__o.dns_type and self.ttl==__o.ttl \
            and (self.priority==__o.priority \
            or all(x in [Number(None), None] for x in [self.priority, __o.priority])) \
            and (self.weight==__o.weight \
            or all(x in [Weight(1), None] for x in [self.weight, __o.weight])) \
            and (self.line==__o.line \
            or all(x in [DNSLine('default'), None] for x in [self.line, __o.line])):
            return True
        else:
            return False


from ddd_objects.domain.value_obj import ExpiredValueObject
from ddd_objects.domain.exception import ParameterError

class Number(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Size(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class ID(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Status(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class SecurityGroupID(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class InstanceType(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Hostname(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class IngressHost(ExpiredValueObject):
    def __init__(self, value):
        self.parts = value.split('.')
        if len(self.parts)>3 or len(self.parts)==1:
            raise ParameterError('Invaild value')
        super().__init__(value, None)
    def get_domain_name(self):
        if len(self.parts)==2:
            return DomainName(self.value)
        elif len(self.parts)==3:
            return DomainName('.'.join(self.parts[-2:]))
    def get_subdomain(self):
        if len(self.parts)==2:
            return Subdomain('@')
        elif len(self.parts)==3:
            return Subdomain(self.parts[0])

class Price(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class ImageID(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class RegionID(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class ZoneID(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class InternetPayType(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class PayType(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class DateTime(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class IP(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class BandWidth(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Password(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Data(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Bool(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Command(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Username(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Port(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Output(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class InstanceID(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class InstanceName(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Endpoint(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Token(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class NodeLabel(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Usage(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class GPUType(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class InstanceTypeStatus(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class InstanceTypeStatusCategory(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Time(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Version(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Name(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)
    def match(self, value):
        return self.value==value

class Type(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Path(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class NodeName(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class NodeType(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class NodeStatus(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class TimeInterval(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Key(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Value(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class KeyType(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Info(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class DomainName(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class RecordID(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Weight(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class DNSType(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class DNSLine(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class SubdomainName(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Subdomain(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)


from typing import List, Optional
from ddd_objects.domain.repository import Repository
from .entity import(
    Condition,
    InstanceTypeUserSetting,
    InstanceTypeWithStatus,
    InstanceUsageInfo,
    InstanceUserSetting,
    InstanceInfo,
    OSSOperationInfo,
    CommandSetting,
    CommandResult
)
from .value_obj import(
    RegionID,
    Name
)
from ..domain.entity import (
    Condition,
    CommandResult,
    DNSRecord,
)
from ..domain.value_obj import (
    Name,
    DomainName,
    RecordID,
)


class VPCRepository(Repository):
    def new_instance(
        self, 
        condition: Condition, 
        setting: InstanceUserSetting
    )->List[InstanceInfo]:
        raise NotImplementedError

    def oss_operate(
        self, 
        instance_infos: List[InstanceInfo],
        oss_operation_info: OSSOperationInfo,
        command_setting: CommandSetting
    )->List[CommandResult]:
        raise NotImplementedError

    def run_command(
        self,
        instance_infos: List[InstanceInfo],
        command_setting: CommandSetting
    )->List[CommandResult]:
        raise NotImplementedError

    def get_instance_by_name(
        self,
        region_id: RegionID,
        name: Name
    )-> List[InstanceInfo]:
        raise NotImplementedError

    def release_instances(self, infos: List[InstanceInfo]):
        raise NotImplementedError

    def get_instance_type_status(
        self, 
        user_settings: List[InstanceTypeUserSetting],
    ) -> List[InstanceTypeWithStatus]:
        raise NotImplementedError

    def get_instance_usage(
        self,
        instance_infos: List[InstanceInfo]
    )->List[InstanceUsageInfo]:
        raise NotImplementedError


class DNSRepository(Repository):

    def create_dns_record(self, record: DNSRecord)->Optional[DNSRecord]:
        raise NotImplementedError

    def save_dns_record(self, record: DNSRecord)->Optional[DNSRecord]:
        raise NotImplementedError

    def delete_dns_record(self, record_id: RecordID)->None:
        raise NotImplementedError

    def get_dns_records(self, domain_name: DomainName)->Optional[List[DNSRecord]]:
        raise NotImplementedError

    def update_dns_record(self, record: Optional[List[DNSRecord]])->Optional[DNSRecord]:
        raise NotImplementedError




import os, json
from typing import List, Optional
from ddd_objects.infrastructure.repository_impl import RepositoryImpl
from ddd_objects.infrastructure.repository_impl import (
    RepositoryImpl, 
    return_list, 
    default_log_fun,
)
from ddd_objects.infrastructure.ao import exception_class_dec
from ddd_objects.domain.exception import NotExistsError, ParameterError, OperationError, ValueError
from ddd_objects.lib import Logger, get_md5
from ..domain.repository import VPCRepository
from .ao import VPCController
from .converter import(
    CommandResultConverter,
    CommandSettingConverter,
    ConditionConverter,
    InstanceInfoConverter,
    InstanceTypeUserSettingConverter,
    InstanceTypeWithStatusConverter,
    InstanceUserSettingConverter,
    OSSOperationInfoConverter
)
from ..domain.entity import(
    CommandResult,
    CommandSetting,
    Condition,
    InstanceInfo,
    InstanceTypeUserSetting,
    InstanceTypeWithStatus,
    InstanceUsageInfo,
    InstanceUserSetting,
    OSSOperationInfo
)
logger = Logger()
logger.set_labels(file_name=__file__)
from ..domain.entity import (
    Condition,
    CommandResult,
    DNSRecord,
)
from ..domain.value_obj import (
    DomainName,
    RecordID,
    Subdomain,
    Usage,
    Name,
    Number,
    Command,
    RegionID,
    Size
)
from .converter import (
    DNSRecordConverter,
    ConditionConverter,
    CommandResultConverter,
)
from ..domain.repository import (
    DNSRepository
)
instance_info_converter = InstanceInfoConverter()
condition_converter = ConditionConverter()
instance_user_setting_converter = InstanceUserSettingConverter()
oss_operation_info_converter = OSSOperationInfoConverter()
command_setting_converter = CommandSettingConverter()
command_result_converter = CommandResultConverter()
instance_type_status_converter = InstanceTypeWithStatusConverter()
instance_type_user_converter = InstanceTypeUserSettingConverter()
dns_record_converter = DNSRecordConverter()


class VPCRepositoryImpl(VPCRepository, RepositoryImpl):
    def __init__(self, ip, port, token, log_func=None):
        self.ao = VPCController(ip, port, token)
        super().__init__(log_func=log_func)

    def _extract_cpu_usage_info(self, result_output:str):
        lines = result_output.split('\n')
        values = [v for v in lines[0].split(' ') if v]
        cpu_num = values[-2].split('\t')[-1].replace('(', '').strip()
        cpu_free = [v for v in lines[3].split(' ') if v][-1]
        _, _, io_out, io_in, _, _ = [v for v in lines[6].split(' ') if v]
        cpu_usage = 1-float(cpu_free)/100
        return int(cpu_num), cpu_usage, float(io_in), float(io_out)

    def _extract_memory_usage_info(self, result_output: str):
        lines = result_output.split('\n')
        _, total, usage, _, _, _, _ = [v for v in lines[1].split(' ') if v]
        return total, usage

    def _extract_flow_info(self, result_output: str):
        lines = result_output.split('\n')
        line = [line for line in lines if line.startswith('eth')][0]
        _, _, _, _, _, _, flow_in, _, flow_out = [v for v in line.split(' ') if v]
        flow_in = flow_in.replace('K', '000')
        flow_out = flow_out.replace('K', '000')
        flow_in = flow_in.replace('M', '000000')
        flow_out = flow_out.replace('M', '000000')
        flow_in = flow_in.replace('G', '000000000')
        flow_out = flow_out.replace('G', '000000000')
        return float(flow_in)/1000, float(flow_out)/1000

    def _extract_disk_info(self, result_output: str):
        lines = result_output.split('\n')
        line = [line for line in lines 
            if line.startswith('/dev/vda') and line.strip().endswith('/')][0]
        _, size, used, avail, usage, _ = [v for v in line.split(' ') if v]
        size = int(size[:-1])
        usage = float(usage[:-1])/100
        return size, usage

    @return_list(log_func=default_log_fun, return_raw=True)
    @exception_class_dec(max_try=1)
    def new_instance(
        self, 
        condition: Condition, 
        setting: InstanceUserSetting,
        force: bool = False
    ) -> List[InstanceInfo]:
        if not force:
            region_id = setting.region_id.get_value()
            name = setting.name.get_value()
            _result = self.ao.get_instance_by_name(region_id, name)
            if _result.succeed:
                infos = _result.get_value()
            else:
                self.log_func('Fail to fetch existing instance infos when creating instance')
                self.log_func(_result.error_traceback)
                return None
            if infos:
                return infos
        condition = condition_converter.to_do(condition)
        setting = instance_user_setting_converter.to_do(setting)
        _result = self.ao.new_instance(condition, setting)
        if _result.succeed:
            instance_infos = _result.get_value()
            if instance_infos:
                return [instance_info_converter.to_entity(info) for info in instance_infos]
            else:
                return None
        else:
            self.log_func('Fail to creating instance')
            self.log_func(_result.error_traceback)
            return None

    @return_list(log_func=default_log_fun, return_raw=True)
    @exception_class_dec()
    def oss_operate(
        self, 
        instance_infos:List[InstanceInfo], 
        oss_operation_info: OSSOperationInfo,
        command_setting: CommandSetting
    ) -> List[CommandResult]:
        instance_infos = [instance_info_converter.to_do(info) for info in instance_infos]
        oss_operation_info = oss_operation_info_converter.to_do(oss_operation_info)
        command_setting = command_setting_converter.to_do(command_setting)
        result = self.ao.oss_operate(instance_infos, oss_operation_info, command_setting)
        if result.succeed:
            return [command_result_converter.to_entity(r) for r in result.get_value()]
        else:
            self.log_func('Fail to operate oss')
            self.log_func(result.error_traceback)
            return None

    @return_list(log_func=default_log_fun, return_raw=True)
    @exception_class_dec()
    def run_command(
        self,
        instance_infos: List[InstanceInfo],
        command_setting: CommandSetting
    ) -> List[CommandResult]:
        instance_infos = [instance_info_converter.to_do(info) for info in instance_infos]
        command_setting = command_setting_converter.to_do(command_setting)
        _result = self.ao.run_command(instance_infos, command_setting)
        if _result.succeed:
            results = _result.get_value()
            return [command_result_converter.to_entity(r) for r in results]
        else:
            self.log_func(f'Fail to run command: {command_setting.command}')
            self.log_func(_result.error_traceback)
            return None

    def _get_instance_by_name(self, region_id: RegionID, name: Name) -> List[InstanceInfo]:
        region_id = region_id.get_value()
        name = name.get_value()
        result = self.ao.get_instance_by_name(region_id, name)
        if result.succeed:
            return [instance_info_converter.to_entity(x) for x in result.get_value()]
        else:
            self.log_func(f'Fail to fetch instance info')
            self.log_func(result.error_traceback)
            return None

    def get_instance_by_name(self, region_id: RegionID, name: Name) -> List[InstanceInfo]:
        key = f'{region_id.get_value()}:{name.get_value()}:instances'
        return self.find_entity_helper(
            self._get_instance_by_name,
            key,
            converter=None,
            force_update=False,
            verbose = False,
            region_id=region_id,
            name = name
        )

    def _get_instance_type_status(
        self, 
        user_settings: List[InstanceTypeUserSetting],
    ) -> List[InstanceTypeWithStatus]:
        user_settings = [instance_type_user_converter.to_do(x) for x in user_settings]
        result = self.ao.get_instance_type_status(user_settings)
        if result.succeed:
            return [instance_type_status_converter.to_entity(x) for x in result.get_value()]
        else:
            self.log_func(f'Fail to fetch instance type status info')
            self.log_func(result.error_traceback)
            return None

    def get_instance_type_status(
        self, 
        user_settings: List[InstanceTypeUserSetting],
    ) -> List[InstanceTypeWithStatus]:
        region_ids = [setting.region_id.get_value() for setting in user_settings]
        zone_ids = [setting.zone_id.get_value() for setting in user_settings]
        instance_types = [setting.instance_type_id.get_value() for setting in user_settings]
        region_ids_string = ''.join(region_ids)
        zone_ids_string = ''.join(zone_ids)
        instance_types_string = ''.join(instance_types)
        key = f'{get_md5(region_ids_string)}:{get_md5(zone_ids_string)}:{get_md5(instance_types_string)}'
        return self.find_entity_helper(
            self._get_instance_type_status,
            key,
            converter=None,
            user_settings=user_settings
        )

    def _get_instance_usage(
        self,
        instance_infos: List[InstanceInfo]
    )->List[InstanceUsageInfo]:
        usage_info = {}
        for info in instance_infos:
            usage_info[info.id.get_value()] = {}
        # cpu usage and io
        command = CommandSetting(Command('iostat'))
        results1 = self.run_command(instance_infos, command)
        # memory
        command = CommandSetting(Command('free -m'))
        results2 = self.run_command(instance_infos, command)
        # flow io
        command = CommandSetting(Command('ifstat'))
        results3 = self.run_command(instance_infos, command)
        set_sampling = True
        # disk
        command = CommandSetting(Command('df -hl'))
        results4 = self.run_command(instance_infos, command)
        for r1, r2, r3, r4 in zip(results1, results2, results3, results4):
            cpu_num, cpu_usage, io_in, io_out = self._extract_cpu_usage_info(r1.output.get_value())
            _id = r1.instance_id.get_value()
            usage_info[_id]['cpu_num'] = cpu_num
            usage_info[_id]['cpu_usage'] = cpu_usage
            usage_info[_id]['io_in'] = io_in
            usage_info[_id]['io_out'] = io_out
            total_mem, usage_mem = self._extract_memory_usage_info(r2.output.get_value())
            _id = r2.instance_id.get_value()
            usage_info[_id]['total_mem'] = total_mem
            usage_info[_id]['usage_mem'] = usage_mem
            flow_in, flow_out = self._extract_flow_info(r3.output.get_value())
            _id = r3.instance_id.get_value()
            usage_info[_id]['flow_in'] = flow_in
            usage_info[_id]['flow_out'] = flow_out
            if flow_in==0 or flow_out==0:
                set_sampling = False
            disk_size, disk_usage = self._extract_disk_info(r4.output.get_value())
            usage_info[_id]['disk_size'] = disk_size
            usage_info[_id]['disk_usage'] = disk_usage
        if not set_sampling:
            command = CommandSetting(
                Command('ifstat -d 1')
            )
            self.run_command(instance_infos, command)
            command = CommandSetting(
                Command('ifstat')
            )
            results3 = self.run_command(instance_infos, command)
            for r3 in results3:
                flow_in, flow_out = self._extract_flow_info(r3.output.get_value())
                _id = r3.instance_id.get_value()
                usage_info[_id]['flow_in'] = flow_in
                usage_info[_id]['flow_out'] = flow_out
        instance_usage_infos = []
        for info in instance_infos:
            id = info.id.get_value()
            instance_usage_infos.append(
                InstanceUsageInfo(
                    instance_id=id,
                    instance_name = info.name,
                    cpu_number= Number(usage_info[id]['cpu_num']),
                    cpu_usage = Usage(usage_info[id]['cpu_usage']),
                    memory_size = Size(usage_info[id]['total_mem']),
                    memory_usage= Size(usage_info[id]['usage_mem']),
                    flow_in = Size(usage_info[id]['flow_in']),
                    flow_out = Size(usage_info[id]['flow_out']),
                    disk_size = Size(usage_info[id]['disk_size']),
                    disk_usage= Size(usage_info[id]['disk_usage']),
                    io_in = Size(usage_info[id]['io_in']),
                    io_out = Size(usage_info[id]['io_out'])
                )
            )
        return instance_usage_infos

    def get_instance_usage(
        self,
        instance_infos: List[InstanceInfo]
    )->List[InstanceUsageInfo]:
        instance_ids = [info.id.get_value() for info in instance_infos]
        instance_ids_string = ''.join(instance_ids)
        key = f'{get_md5(instance_ids_string)}:instance_usage'
        return self.find_entity_helper(
            self._get_instance_usage,
            key=key,
            converter=None,
            instance_infos=instance_infos
        )

    def release_instances(self, infos: List[InstanceInfo]):
        infos = [instance_info_converter.to_do(info) for info in infos]
        result = self.ao.release_instances(infos)
        if result.succeed:
            return True
        else:
            instance_names_str = ','.join([info.name for info in infos])
            self.log_func(f'Fail to release instances: {instance_names_str}')
            self.log_func(result.error_traceback)
            return False


class DNSRepositoryImpl(DNSRepository, RepositoryImpl):
    def __init__(self, ip, port, token, log_func=None):
        self.ao = VPCController(ip, port, token)
        super().__init__(log_func=log_func)

    def create_dns_record(self, record: DNSRecord)->Optional[DNSRecord]:
        record = dns_record_converter.to_do(record)
        result = self.ao.create_dns_record(record)
        if result.succeed:
            return dns_record_converter.to_entity(result.get_value())
        else:
            dn = record.domain_name
            subdomain = record.subdomain
            self.log_func(f'Fail to create dns record: {subdomain}/{dn}')
            self.log_func(result.error_traceback)
            return None

    def save_dns_record(self, record: DNSRecord)->Optional[DNSRecord]:
        domain_name = record.domain_name
        existing_records = self.get_dns_records(domain_name)
        for r in existing_records:
            if r.domain_name==record.domain_name and r.subdomain==record.subdomain \
                and r.value==record.value:
                if r==record:
                    return r
                else: 
                    record.id = r.id
                    return self.update_dns_record(record)
        return self.create_dns_record(record)

    def delete_dns_record(self, record_id: RecordID)->None:
        record_id = record_id.get_value()
        result = self.ao.delete_dns_record(record_id)
        if result.succeed:
            return None
        else:
            self.log_func(f'Fail to delete dns record: {record_id}')
            self.log_func(result.error_traceback)
            return None

    def _get_dns_records(self, domain_name: DomainName)->Optional[List[DNSRecord]]:
        domain_name = domain_name.get_value()
        result = self.ao.get_dns_records(domain_name)
        if result.succeed:
            return [dns_record_converter.to_entity(x) for x in result.get_value()]
        else:
            self.log_func(f'Fail to fetch dns record: {domain_name}')
            self.log_func(result.error_traceback)
            return None

    def get_dns_records(self, domain_name: DomainName)->Optional[List[DNSRecord]]:
        key = f'{domain_name}:dns_records'
        return self.find_entity_helper(
            self._get_dns_records,
            key=key,
            converter=None,
            domain_name=domain_name
        )

    def update_dns_record(self, record: DNSRecord)->Optional[DNSRecord]:
        records = self.get_dns_records(record.domain_name)
        for r in records:
            if record==r:
                return r
        record = dns_record_converter.to_do(record)
        result = self.ao.update_dns_record(record)
        if result.succeed:
            return dns_record_converter.to_entity(result.get_value())
        else:
            dn = record.domain_name
            sub = record.subdomain
            self.log_func(f'Fail to update dns record: {dn}/{sub}')
            self.log_func(result.error_traceback)
            return None
